import datetime
import os
from os.path import expanduser


class ServiceType:
    TerraForm = "TerraForm"


class SidecarFilesStructure:
    FTP_FOLDER = "/var/ftp"

    @staticmethod
    def get_ftp_folder(app_name: str):
        return os.path.join(SidecarFilesStructure.FTP_FOLDER, app_name)

    # @staticmethod
    # def get_service_folder(name: str):
    #     return os.path.join(SidecarFilesStructure().get_ftp_folder(name))


# class ServiceKind:
#     TerraForm = "TerraForm"


class Signals:
    ON_INSTANCE_UPDATE_STATUS = "ON_INSTANCE_UPDATE_STATUS"
    ON_APPS_DEPLOYMENT_COMPLETE = "ON_APPS_DEPLOYMENT_COMPLETE"


class AppNetworkStatus:
    PENDING = "pending"
    TESTING_PUBLIC_NETWORK = "testing_public_network"
    TESTING_PRIVATE_NETWORK = "testing_private_network"
    COMPLETED = "completed"
    ERROR = "error"

    @staticmethod
    def is_end_status(status: str) -> bool:
        return status in [AppNetworkStatus.COMPLETED, AppNetworkStatus.ERROR]

    @staticmethod
    def passed_internal_network_test(status: str) -> bool:
        return status in [AppNetworkStatus.TESTING_PUBLIC_NETWORK, AppNetworkStatus.COMPLETED]


class DateTimeProvider:
    def get_current_time_utc(self):
        return datetime.datetime.now(datetime.timezone.utc)


class AppInstanceConfigStatus:
    PENDING = "pending"
    COMPLETED = "completed"
    ERROR = "error"

    END_STATUSES = [COMPLETED, ERROR]

    @staticmethod
    def is_end_status(status: str) -> bool:
        return status in AppInstanceConfigStatus.END_STATUSES


class SandboxDeploymentEndStatus:
    COMPLETED = "completed"
    ERROR = "error"


def _get_tag_name(tag_name: str):
    return 'colony-{}'.format(tag_name)


def get_app_selector(app_name: str):
    return _get_tag_name('selector-{app_name}'.format(app_name=app_name))


class Const:
    EXTERNAL_ELB_DNS_NAME = _get_tag_name('external-elb-dsn')
    APP_STATE_KEY_VALUE_SEPARATOR = ":"
    APP_NAME_TAG = _get_tag_name('app-name')
    APP_STATUS_TAG = _get_tag_name('app-status')
    APPS = _get_tag_name('apps')
    HEALTH_CHECK_STATUS = _get_tag_name("health-check-status")
    PRODUCTION_ID_TAG = _get_tag_name('production-id')
    SANDBOX_ID_TAG = _get_tag_name('sandbox-id')
    SANDBOX_START_TIME = 'start-time'
    SANDBOX_START_TIME_v2 = _get_tag_name('start-time')

    SANDBOX_DEPLOYMENT_END_STATUS = 'deployment-end-status'
    SANDBOX_DEPLOYMENT_END_STATUS_v2 = _get_tag_name('deployment-end-status')
    CSV_TAG_VALUE_SEPARATE = ","

    K8S_SIDECAR_SERVICE = 'sidecar-service'
    K8S_SIDECAR_APP_SELECTOR = get_app_selector(K8S_SIDECAR_SERVICE)

    AWS_SIDECAR_APP_NAME = 'colonySidecar'

    QUALY_SERVICE_NAME = 'colonyDebugService'

    SIDECAR_HOME_FOLDER = os.path.join(expanduser("~"), 'sidecar')

    INSTANCELOGICALID = _get_tag_name('instance-logical-id')
    EXTERNAL_PORTS = _get_tag_name('external-ports')
    INTERNAL_PORTS = _get_tag_name('internal-ports')
    MAIN_ALB_SG = 'MainALBSG'

    SERVICE_EXECUTION_POD = _get_tag_name('service-execution-pod')

    @staticmethod
    def get_config_file():
        return '{HOME_FOLDER}/config.json'.format(HOME_FOLDER=Const.SIDECAR_HOME_FOLDER)

    @staticmethod
    def get_log_file():
        return '{HOME_FOLDER}/log.txt'.format(HOME_FOLDER=Const.SIDECAR_HOME_FOLDER)

    @staticmethod
    def get_kubernetes_token_file_path() -> str:
        return '/var/run/secrets/kubernetes.io/serviceaccount/token'

    @staticmethod
    def get_app_folder(app_name: str) -> str:
        return '{HOME_FOLDER}/{APP_NAME}'.format(HOME_FOLDER=Const.SIDECAR_HOME_FOLDER,
                                                 APP_NAME=app_name)

    @staticmethod
    def get_app_log_file(app_name: str) -> str:
        return '{APP_FOLDER}/log.txt'.format(APP_FOLDER=Const.get_app_folder(app_name=app_name),
                                             app_name=app_name)

    @staticmethod
    def get_health_check_file(app_name: str, script_name: str) -> str:
        return '{APP_FOLDER}/{SCRIPT_NAME}'.format(
            APP_FOLDER=Const.get_app_folder(app_name=app_name), SCRIPT_NAME=script_name)



from logging import Logger

from sidecar.const import Const


class AwsTagHelper:
    SpotFleetIdTag = "aws:ec2spot:fleet-request-id"
    AutoScalingGroupNameTag = "aws:autoscaling:groupName"

    @staticmethod
    def get_tags(resource, logger: Logger) -> {}:
        import time
        for counter in range(60):
            tags = resource.tags
            if not tags:
                resource.reload()
            if not tags:
                logger.warning("Could not get tags retrying - retry {retry} of 60".format(retry=counter))
                time.sleep(5)
                continue
            else:
                break

        if not tags:
            logger.exception("couldn't get tags from resource")

        return {tag['Key']: tag['Value'] for tag in tags}

    @staticmethod
    def safely_get_tag(resource, tag_name: str, logger: Logger, value_if_missing: str = None) -> str:
        return AwsTagHelper.get_tags(resource=resource, logger=logger).get(tag_name, value_if_missing)

    @staticmethod
    def create_tag(key, value):
        return {'Key': key, 'Value': value}

    @staticmethod
    def get_apps_status_map(instance, logger: Logger) -> {}:
        apps_status_tag_value = AwsTagHelper.safely_get_tag(instance, Const.APP_STATUS_TAG, logger)
        return AwsTagHelper.parse_apps_status_tag_value(apps_status_tag_value)

    @staticmethod
    def parse_apps_status_tag_value(apps_status_value: str) -> {}:
        app_to_status_map = dict()
        if apps_status_value:
            all_statuses = apps_status_value.split(Const.CSV_TAG_VALUE_SEPARATE)
            app_to_status_map = dict(state.split(Const.APP_STATE_KEY_VALUE_SEPARATOR) for state in all_statuses)
        return app_to_status_map

    @staticmethod
    def format_apps_status_tag_value(app_to_status_map: {}) -> str:
        return Const.CSV_TAG_VALUE_SEPARATE.join(
            ["{KEY}{SEP}{VALUE}".format(KEY=app_name, SEP=Const.APP_STATE_KEY_VALUE_SEPARATOR, VALUE=app_status)
             for app_name, app_status in app_to_status_map.items()])



import boto3
from botocore.config import Config
from logging import Logger

from sidecar.model.objects import AwsSidecarConfiguration
from sidecar.utils import Utils


class AwsSession:
    def __init__(self, region_name: str, aws_config: AwsSidecarConfiguration, logger: Logger):

        self.logger = logger
        self.aws_config = aws_config
        self._stack_name = None

        config_dict = {'connect_timeout': 20, 'read_timeout': 20, 'max_pool_connections': 100,
                       'retries': {'max_attempts': 2}}
        self._config = Config(**config_dict)
        self._region_name = region_name

        self._init_session()

        self.cf_client = self._get_client('cloudformation', self._config)

        if aws_config:
            try:
                Utils.wait_for(func=lambda: self._test_credentials() is not None,
                               interval_sec=1,
                               max_retries=5,
                               error='Credentials service is not responding')
            except Exception:
                self.logger.exception(msg="Failed to access credentials service after 5 times")
                return

        self.ec2_client = self._get_client('ec2', self._config)

        self.ec2_resource = self._get_resource('ec2', self._config)
        self._autoscalling_client = self._get_client('autoscaling', self._config)
        self._elb_client = self._get_client('elb', self._config)
        self.cloudwatch_client = self._get_client('logs', self._config)

    def _init_session(self):
        self.session = boto3.Session()

    def get_ec2_client(self):
        return self.ec2_client

    def get_cf_client(self):
        return self.cf_client

    def get_ec2_resource(self):
        return self.ec2_resource

    def get_cloudwatch_client(self):
        return self.cloudwatch_client

    def get_autoscaling_client(self):
        return self._autoscalling_client

    def get_elb_client(self):
        return self._elb_client

    def _get_client(self, client: str, config: Config):
        return self.session.client(client, region_name=self._region_name, config=config)

    def _get_resource(self, resource: str, config: Config):
        return self.session.resource(resource, region_name=self._region_name, config=config)

    def get_dynamo_resource(self, default_region: str = None):
        region = self._region_name

        if default_region:
            region = default_region

        return self.session.resource('dynamodb', region_name=region, config=self._config)

    def _test_credentials(self):
        try:
            response = self.cf_client.describe_stacks(StackName=self.stack_name, NextToken='0')
            self.logger.info("test credentials Success")
            return response
        except Exception:
            self.logger.info("test credentials FAIL", exc_info=True)
            return None

    @property
    def stack_name(self):
        if not self._stack_name:
            if self.aws_config.production_id:
                self._stack_name = f'production-{self.aws_config.production_id}-sandbox-{self.aws_config.sandbox_id}'
            else:
                self._stack_name = f'sandbox-{self.aws_config.sandbox_id}'

        return self._stack_name
